/*  &copy; Copyright 2012 by Al Williams (al.williams@awce.com) 
    Distributed under the terms of the GNU Lesser General Public License */

/*! 
 * \file GP3.cpp
 * \author Al Williams
 * \date 20 Nov 2012
 * \brief Main library source for GP3/Arduino
 *
 * The GP3/Arduino is a coprocessor for the 
 * Arduino that adds additional analog and
 * digital I/O easily via two Arduino pins

    This file is part of GP3/Arduino.

    GP3/Arduino is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.


 */

/*! \mainpage GP3/Arduino
 * &copy; Copyright 2012 by Al Williams (al.williams@awce.com)
 * Find out more at http://www.awce.com/gp3.htm
 * Distributed under the terms of the GNU General Public License
 * \section license License

    This file is part of GP3/Arduino.

    GP3/Arduino is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    * \section about About
 * The GP3/Arduino is a coprocessor for the 
 * Arduino that adds additional analog and
 * digital I/O easily via two Arduino pins.
 * It works with the SoftwareSerial device
 * or any other Stream-based serial port

 * There are two versions of this library.
 * GP3 uses the SoftwareSerial library directly or you
 * can provide a port object using any compatible library.
 * To use, import as usual. You also need to import 
 * SoftwareSerial.h at the top of your sketch using and include
 * statement.
 *
 * Construct a GP3 object with the receive and 
 * transmit pins you want to use and you 
 * are ready to go.

 * The other version (GP3PORT) does not include SoftwareSerial
 * which should make your sketches smaller if you are using
 * a different port library. However, then you must provide
 * a port object. The regular GP3 library will work either way
 * GP3PORT only works with a port object you supply
 * (even if it is a SoftwareSerial port).
 */

#include <GP3.h>

//! Timeout in ms -- if the GP3 fails to respond in this time, an error occurs
#define TIMEOUT 100 

//! Construct a GP3 object with a SoftwareSerial port
//! NOTE: Not available in port version
GP3::GP3(int rx, int tx) 
    {
      timeout=TIMEOUT;
      port=sport=new SoftwareSerial(rx,tx);
      // baud rate must be 57600
      sport->begin(57600);
      // no error, default analog channel
      err=0;
      achannel=0;
    }

//! Construct a GP3 object with any port
//! Note must be 57600 baud
//! This is the port version's only constructor
GP3::GP3(Stream &_port)
{
  timeout=TIMEOUT;
  port=&_port;
  err=0;
  achannel=0;
  sport=NULL;  // remember this isn't the default
}

//! Clean up if using the default port
//! If not using the default port you are
//! on your own.
GP3::~GP3()
{
  if (sport) delete sport;
}

//! Returns non-zero if an error has occurred
//! Never automatically cleared after construction
int GP3::getError(void)
{
  return err;
}

//! Clear any errors
void GP3::clearError(void)
{
  err=0;
}


//! Set a communication timeout (default 100mS)
void GP3::setTimeout(int _timeout)
{
  timeout=_timeout;
}

//! Get current communication timeout
int GP3::getTimeout(void)
{
  return timeout;
}

//! Call the soft port's listen function
//! If using your own port, always returns
//! false even if you are using a soft port
//! Not available in the port version
bool GP3::listen(void)
{
  return sport?sport->listen():false;
}

//! Call the soft port's isListening function
//! If using your own port, always returns
//! false even if you are using a soft port
//! Not available in the port version
bool GP3::isListening()
{
  return sport?sport->isListening():false;
}



//! Read a byte from the GP3 with a timeout
//! You usually do not call this directly
//! Unless you are using repeat
int GP3::readByte(void)
{
  int rv;
  long mend=(long)millis();  // mark time
  mend+=timeout;  // compute bail time
  do 
    {
      rv=port->read();  // get a char
      // check for timeout
      if (rv==-1 && timeout!=-1 && (long)millis()-mend>=0) break;
    } while (rv==-1);
  if (rv==-1) err=1;  // note error
  return rv;
}

//! Set the GP3 argument -- rarely used from user programs
//! Unless you are directly manipulating the GP3 protocol
void GP3::setarg(unsigned int arg)
{
  if (arg<=255)
    {
      port->write(6+(arg&1));
      port->write(0x80+(arg&1));
    }
  else
    {
      int code=8+(arg&1);
      port->write(code+((arg>>8)&1));
      port->write(0x80+(arg>>9));
      port->write(0x80+((arg&0xFF)>>1));
    }
}

//! Read a word from the GP3. Rarely called from
//! a user program unless using repeat
unsigned int GP3::readWord(void)
{
  unsigned int rv;
  rv=readByte()<<8;
  return rv+readByte();
}

//! Set the LED on (true) or off (false)
void GP3::setLED(boolean onoff)
{
  port->write(0xC+(onoff&1));
}

//! Select an A/D channel (default=0)
//! All subsequent calls to a2d will
//! use this channel
void GP3::a2dChan(int chan)
{
  achannel=chan;
  a2d();
}


//! Get the analog value (0-1023) on the current
//! channel
unsigned int GP3::a2d(void)
{
  port->write(0x30+((achannel&7)<<1));
  return readWord();
}

//! Configure A/D convert for right justification
//! (default) and voltage reference
//! Rarely used; see GP3 manual for details
void GP3::a2dConf(bool rightjustify, int cselect)
{
  int cmd=0x80+(cselect&7);
  if (rightjustify) cmd|=0x40;
  port->write(0x31);
  port->write(cmd);
}

//! Set a digital output pin high
void GP3::high(int pin)
{
  port->write(0x11+(pin<<1));
}

//! Set a digital output pin low
void GP3::low(int pin)
{
  port->write(0x10+(pin<<1));
}

//! Toggle a digital output pin (set high if low, low if high)
void GP3::toggle(int pin)
{
  port->write(0x20+(pin<<1));
}

//! Get an input from a digital pin
int GP3::input(int pin)
{
  port->write(0x21+(pin<<1));
}

//! Set the port I/O bits all at once
void GP3::setTris(int tris)
{
  port->write(2+(tris &1));
  port->write(0x80+(tris>>1));
}

//! Get the port I/O bits
int GP3::getTris(void)
{
  port->write(0x56);
  return readByte();
}

//! Set all output bits at once
void GP3::setPins(int byte)
{
  port->write(4+(byte&1));
  port->write(0x80+(byte>>1));
}

//! Get all input bits at once
int GP3::getPins(void)
{
  port->write(0x57);
  return readByte();
}

//! Write to EEPROM
void GP3::writeEE(int address, int byte)
{
  setarg(address);
  port->write(0x50+(byte&1));
  port->write(0x80+(byte>>1));
}

//! Read from EEPROM
int GP3::readEE(int address)
{
  setarg(address);
  port->write(0x52);
  return readByte();
}

//! Generate PWM
//! Use chan=-1 for hardware channel
void GP3::pwm(int chan, int dc, int dfreq)
{
  setarg(dfreq);
  port->write((chan==-1)?0x54:(0x60+(chan<<1))+(dc&1));
  port->write(0x80+(dc>>1));
}

//! Count events on a pin over a specified number of milliseconds
unsigned int GP3::count(int pin, unsigned int ms)
{
  setarg(ms);
  port->write(0x40+(pin<<1));
  return readWord();
}

//!  Perform an RC timing algorithm on a pin (see GP3 manual)
unsigned int GP3::rcTime(int pin, int state)
{
  port->write(0xe);
  port->write(0xA0+(state&1)+(pin<<1));
  return readWord();
}

//! Measure a pulse width
unsigned int GP3::pulseIn(int pin, int state)
{
  port->write(0x70+(state&1)+(pin<<1));
  return readWord();
}

//! Create a pulse of a given duration
void GP3::pulseOut(int pin, int duration)
{
  setarg(duration);
  port->write(0x41+(pin<<1));
}

//! Output a frequency
void GP3::freq(int pin, int ms, int frq)
{
  setarg(ms);
  port->write(0xE);
  port->write(0x80+(frq&1)+(pin<<1));
  port->write(0x80+(frq>>8));
  port->write(0x80+((frq&0xff)>>1));
}

//! Set hardware counter prescaler and source
//! Prescaler can be 1, 2, 4, or 8
void GP3::setCounter(int prescale, int ext)
{
  int ps=0;
  switch (prescale)
    {
    case 1:  ps=0; break;
    case 2:  ps=2; break;
    case 4:  ps=4; break;
    case 8:  ps=6; break;
    }
  port->write(0xE);
  port->write(0x90+ps+(ext&1));
}

//! Get hardware counter value
unsigned int GP3::count(void)
{
  port->write(0x5a);
  return readWord();
}

//! Reset everything
void GP3::resetAll(void)
{
  port->write(0x5c);  // reset in case of multibyte command
  port->write(0x5c);  // really reset
}

//! Warm reset
void GP3::reset(void)
{
  port->write((uint8_t)0);
}

//! The repeat command allows the next command
//! to repeat over and you get the 2nd and later
//! results using readword() or readbyte()
void GP3::repeat(int count)
{
  setarg(count-1);
  port->write(0x5b);
}

//! Increment the arg
void GP3::argInc(void)
{
  port->write(0x59);
}

//! Decrement the arg
void GP3::argDec(void)
{
  port->write(0x58);
}

//! Check the GP3 status. Should return 51
unsigned int GP3::check(void)
{
  int rv;
  port->write(1);
  rv=readByte();
  err=rv!=51;
  return rv;
}

    

 

  



